"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = __importDefault(require("child_process"));
const fs_1 = require("fs");
const git_log_parser_1 = __importDefault(require("git-log-parser"));
const path_1 = __importDefault(require("path"));
const semver_1 = __importDefault(require("semver"));
const stream_1 = require("stream");
const util_1 = require("util");
const get_conventional_bump_1 = require("./utils/get-conventional-bump");
const get_packages_in_topological_order_1 = require("./utils/get-packages-in-topological-order");
const max_increment_1 = require("./utils/max-increment");
const execFile = (0, util_1.promisify)(child_process_1.default.execFile);
const LAST_BUMP_COMMIT_MESSAGE = process.env.LAST_BUMP_COMMIT_MESSAGE || 'chore(ci): bump packages';
function shouldSkipPackage(packageName) {
    const skippedPackages = (process.env.SKIP_BUMP_PACKAGES || '').split(',');
    const shouldSkip = skippedPackages.includes(packageName);
    console.log('Should skip', packageName, {
        packages: skippedPackages,
        shouldSkip,
    });
    return shouldSkip;
}
async function main() {
    try {
        await fs_1.promises.stat('./package.json');
        await fs_1.promises.stat('.git');
    }
    catch (err) {
        if (err.code === 'ENOENT') {
            throw new Error('this command can only be run from the root of a monorepo');
        }
        throw err;
    }
    const monorepoRootPath = process.cwd();
    const packages = await (0, get_packages_in_topological_order_1.getPackagesInTopologicalOrder)(monorepoRootPath);
    const newVersions = {};
    const range = await getRangeFromLastBump();
    for (const packageInfo of packages) {
        await processPackage(path_1.default.relative(monorepoRootPath, packageInfo.location), newVersions, { range });
    }
    await execFile('npm', ['install', '--package-lock-only']);
}
main().catch((err) => process.nextTick(() => {
    throw err;
}));
async function getCommits({ path, range }) {
    const commits = [];
    const stream = git_log_parser_1.default
        .parse({
        _: [range, path].filter(Boolean),
    })
        .pipe(new stream_1.PassThrough({ objectMode: true }));
    for await (const commit of stream) {
        commits.push(commit);
    }
    return commits;
}
function updateDeps(packageJson, newVersions) {
    console.debug(`[${packageJson.name}]`, 'updateDeps', newVersions);
    const newPackageJson = JSON.parse(JSON.stringify(packageJson));
    let inc;
    for (const sectionName of [
        'dependencies',
        'devDependencies',
        'peerDependencies',
        'optionalDependencies',
    ]) {
        const dependenciesSection = newPackageJson[sectionName];
        if (!dependenciesSection) {
            continue;
        }
        for (const [depName, { version, bump: depInc }] of Object.entries(newVersions)) {
            if (!dependenciesSection[depName]) {
                console.debug(`[${packageJson.name}]`, 'updateDeps', sectionName, depName, 'skipping');
                continue;
            }
            console.debug(`[${packageJson.name}]`, 'updateDeps', sectionName, depName, '->', version);
            const oldDepRange = dependenciesSection[depName];
            const newDepRange = /^\d\.\d\.\d/.exec(oldDepRange)
                ? version
                : oldDepRange === '*'
                    ? '*'
                    : oldDepRange.startsWith('~')
                        ? `~${version}`
                        : `^${version}`;
            dependenciesSection[depName] = newDepRange;
            const oldInc = inc;
            inc =
                sectionName === 'devDependencies'
                    ? (0, max_increment_1.maxIncrement)('patch', inc)
                    : (0, max_increment_1.maxIncrement)(depInc, inc);
            console.debug(`[${packageJson.name}]`, 'inc', sectionName === 'devDependencies', {
                depInc,
                oldInc,
                newInc: inc,
            });
        }
    }
    if (inc) {
        newPackageJson.version = semver_1.default.inc(packageJson.version, inc);
    }
    console.debug(`[${packageJson.name}]`, 'new package json', JSON.stringify(newPackageJson));
    return newPackageJson;
}
async function bumpVersionBasedOnCommits(packagePath, oldVersion, options) {
    const commits = await getCommits({
        path: packagePath,
        range: options.range,
    });
    console.info(`[${packagePath}]`, commits.length, 'commits found in range', commits.length ? '' : '.. skipping');
    if (!commits.length) {
        return oldVersion;
    }
    let inc = 'patch';
    for (const commit of commits) {
        inc = (0, max_increment_1.maxIncrement)(inc, (0, get_conventional_bump_1.getConventionalBump)(commit));
    }
    const newVersion = semver_1.default.inc(oldVersion, inc);
    console.info(`[${packagePath}]`, 'bumping', oldVersion, 'to', newVersion, `(${inc})`);
    return newVersion;
}
async function getRangeFromLastBump() {
    const allCommits = await getCommits({
        path: '.',
    });
    console.info('total commits found', allCommits.length);
    const lastBumpCommit = allCommits.find((c) => c.subject.startsWith(LAST_BUMP_COMMIT_MESSAGE));
    console.info('lastBumpCommit', lastBumpCommit
        ? `${lastBumpCommit.commit.long} ${lastBumpCommit.subject}`
        : '-');
    const range = lastBumpCommit
        ? `${lastBumpCommit.commit.long}...HEAD`
        : undefined;
    console.info('range:', range);
    return range;
}
async function processPackage(packagePath, newVersions, options) {
    const packageJsonPath = path_1.default.join(packagePath, 'package.json');
    const origPackageJsonString = await fs_1.promises.readFile(packageJsonPath, 'utf-8');
    const packageJson = JSON.parse(origPackageJsonString);
    const packageJsonAfterDepBump = updateDeps(packageJson, newVersions);
    let newVersion = packageJson.version;
    if (!shouldSkipPackage(packageJson.name)) {
        const conventionalVersion = await bumpVersionBasedOnCommits(packagePath, packageJson.version, options);
        newVersion = semver_1.default.gt(conventionalVersion, packageJsonAfterDepBump.version)
            ? conventionalVersion
            : packageJsonAfterDepBump.version;
        if (semver_1.default.gt(newVersion, packageJson.version)) {
            newVersions[packageJson.name] = {
                version: newVersion,
                bump: semver_1.default.diff(newVersion, packageJson.version),
            };
        }
    }
    const newPackageJson = { ...packageJsonAfterDepBump, version: newVersion };
    if (!(0, util_1.isDeepStrictEqual)(newPackageJson, packageJson)) {
        const trailingSpacesMatch = /}(\s*)$/.exec(origPackageJsonString);
        const trailingSpaces = (trailingSpacesMatch && trailingSpacesMatch[1]) || '';
        await fs_1.promises.writeFile(packageJsonPath, JSON.stringify(newPackageJson, null, 2) + trailingSpaces, 'utf-8');
    }
}
//# sourceMappingURL=bump-packages.js.map